from .exceptions import QueryError

REPR_OUTPUT_SIZE = 20


def combine_list(a,b):
    if isinstance(a,(set,tuple,list)):
        a = list(a)
    else:
        a = [a]
    if isinstance(b,(set,tuple,list)):
        b = list(b)
    else:
        b = [b]
    a.extend(b)
    return a

def combine_dicts(a, b, op=combine_list):
    z = a.copy()
    z.update(b)
    z.update([(k, op(a[k], b[k])) for k in set(b) & set(a)])
    doc_type = z.get('_doc_type')
    if isinstance(doc_type,list):
        doc_type = set(doc_type)
        if len(doc_type) == 1:
            doc_type = doc_type[0]
            z['_doc_type'] = doc_type
    return z


class QuerySetMixin(object):
    query_type = None

    def __init__(self,doc_class,q=None,parent_q=None):
        self.parent_q=parent_q
        self._result_cache = None
        self._doc_class = doc_class
        self.q = q
        self.evaluated = False
        self._db = self._doc_class.get_db()
        if q and parent_q:
            self.q = self.combine_qs()

    def combine_qs(self):
        return combine_dicts(self.parent_q, self.q)

    def prepare_filters(self):
        filter_list = []
        for k,v in list(self.q.items()):
            if isinstance(v,list):
                for index_v in v:
                    filter_list.append(self._doc_class.get_index_name(k,index_v))
            else:
                filter_list.append(self._doc_class.get_index_name(k,v))
        return filter_list

    def __len__(self):
        self._fetch_all()
        return len(self._result_cache)
           
    def __repr__(self):  # pragma: no cover
        data = list(self[:REPR_OUTPUT_SIZE + 1])
        if len(data) > REPR_OUTPUT_SIZE:
            data[-1] = "...(remaining elements truncated)..."
        return repr(data)   
    
    def __iter__(self):
        self._fetch_all()
        return iter(self._result_cache)
    
    def _fetch_all(self):
        if self._result_cache is None:
            self._result_cache = list(self.evaluate())
        
    def count(self):
        return len(list(self.evaluate()))
            
    def __bool__(self):
        self._fetch_all()
        return bool(self._result_cache)
    
    def __getitem__(self,index):
        if self._result_cache is not None:
            return self._result_cache[index]
        else:
            self._fetch_all()
            return self._result_cache[index]
    
    def evaluate(self):
        raise NotImplementedError


class QuerySet(QuerySetMixin):

    def filter(self,q):
        return QuerySet(self._doc_class,q,self.q)

    def get(self,q):
        qs = QuerySet(self._doc_class,q,self.q)
        if len(qs) > 1:
            raise QueryError('This query should return exactly one result. Your query returned {0}'.format(len(qs)))
        if len(qs) == 0:
            raise QueryError('This query did not return a result.')
        return qs[0]

    def evaluate(self):
        filters_list = self.prepare_filters()
        return self._doc_class.get_db().evaluate(filters_list,self._doc_class)


class QueryManager(object):
    
    def __init__(self,cls):
        self._doc_class = cls
        
        self.filter = QuerySet(self._doc_class).filter
        self.get = QuerySet(self._doc_class).get



import unittest
from kev.loading import KevHandler
from envs import env

kev_handler = KevHandler({
    's3':{
        'backend':'kev.backends.s3.db.S3DB',
        'connection':{
            'bucket':env('S3_BUCKET_TEST'),
            'indexer':{
                'host':env('REDIS_HOST_TEST'),
                'port':env('REDIS_PORT_TEST'),
            }
        }
    },
    'redis': {
        'backend': 'kev.backends.redis.db.RedisDB',
        'connection': {
            'host': env('REDIS_HOST_TEST'),
            'port': env('REDIS_PORT_TEST'),
        }
    }
})

class KevTestCase(unittest.TestCase):

    def tearDown(self):
        for db_label in list(kev_handler._databases.keys()):
            kev_handler.get_db(db_label).flush_db()


"""
Redis document property classes. Much of the date and time property
code was borrowed or inspired by Benoit Chesneau's CouchDBKit library.
"""

import datetime
import time

from decimal import Decimal

import six

from kev.exceptions import ValidationException
from .validators import (RequiredValidator,StringValidator,
                         MaxLengthValidator,MinLengthValidator,
                         IntegerValidator,MaxValueValidator,
                         MinValueValidator,FloatValidator,
                         DateValidator,DateTimeValidator,BooleanValidator
                         )
import collections

class VariableMixin(object):
    def get_validators(self):
        if self.required:
            self.validators.insert(0,RequiredValidator())
    
class CharVariableMixin(VariableMixin):
    
    def get_validators(self):
        VariableMixin.get_validators(self)
        self.validators.append(StringValidator())
        if self.kwargs.get('min_length'):
            self.validators.append(MinLengthValidator(
                self.kwargs.get('min_length')))
            
        if self.kwargs.get('max_length'):
            self.validators.append(MaxLengthValidator(
                self.kwargs.get('max_length')))
        
class NumericVariableMixin(VariableMixin):    
    def get_validators(self):
        VariableMixin.get_validators(self)
        if self.kwargs.get('max_value'):
            self.validators.append(MaxValueValidator(
                self.kwargs.get('max_value')))
            
        if self.kwargs.get('min_value'):
            self.validators.append(MinValueValidator(
                self.kwargs.get('min_value')))
            
class IntegerVariableMixin(NumericVariableMixin):
    
    def get_validators(self):
        super(IntegerVariableMixin,self).get_validators()
        self.validators.insert(0,IntegerValidator())
            
class FloatVariableMixin(NumericVariableMixin):
    
    def get_validators(self):
        super(FloatVariableMixin,self).get_validators()
        self.validators.insert(0, FloatValidator())

      
class DateMixin(VariableMixin):
    
    def get_validators(self):
        super(DateMixin,self).get_validators()
        self.validators.insert(0,DateValidator())
        
class DateTimeMixin(VariableMixin):
    
    def get_validators(self):
        super(DateTimeMixin,self).get_validators()
        self.validators.insert(0,DateTimeValidator())
        

class BooleanMixin(VariableMixin):

    def get_validators(self):
        self.validators = [BooleanValidator()]


class BaseProperty(VariableMixin,object):
    
    def __init__(
        self,
        default_value=None,
        required=True,
        index=False,
        unique=False,
        validators=[],
        verbose_name=None,
        **kwargs
        ):
        self.default_value = default_value
        self.required = required
        self.index = index
        self.unique = unique
        if unique:
            self.index = True
        self.kwargs = kwargs
        self.validators = list()
        self.get_validators()
        self.validators = set(self.validators)
        if verbose_name:
            self.verbose_name = verbose_name
            
    def validate(self,value,key):
        if not value and type(self.get_default_value()) != type(None):
            value = self.get_default_value()
        for i in self.validators:
            i.validate(value,key)
            
    def get_default_value(self):
        """ return default value """
        default = self.default_value
        if isinstance(default, collections.Callable):
            default = default()
        return default
    
    def get_db_value(self,value):
        return value
    
    def get_python_value(self,value):
        return value
    
class CharProperty(CharVariableMixin,BaseProperty):
    
    def get_db_value(self, value):
        return str(value)
    
    def get_python_value(self, value):
        if not value:
            return None
        return str(value)

class IntegerProperty(IntegerVariableMixin,BaseProperty):
    
    def get_db_value(self, value):
        return int(value)
    
    def get_python_value(self, value):
        if not value:
            return None
        return int(value)
    
    
class FloatProperty(FloatVariableMixin,BaseProperty):
    
    def get_db_value(self, value):
        return float(value)
    
    def get_python_value(self, value):
        if not value:
            return None
        return float(value)


class BooleanProperty(BooleanMixin,BaseProperty):

    def get_db_value(self,value):
        return bool(value)

    def get_python_value(self,value):
        if not value:
            return False
        try:
            BooleanValidator().validate(value,'boolean')
        except ValidationException:
            return value
        return bool(value)


class DateProperty(DateMixin,BaseProperty):

    def __init__(
        self, 
        default_value=None, 
        required=True, 
        validators=[], 
        verbose_name=None, 
        auto_now=False,
        auto_now_add=False,
        **kwargs):
        
        super(DateProperty,self).__init__(default_value=default_value,
                required=required, validators=validators, 
                verbose_name=verbose_name, **kwargs)
        self.auto_now = auto_now
        self.auto_now_add = auto_now_add
                
    def now(self):
        return datetime.datetime.now().date()
    
    def get_default_value(self):
        default = self.default_value
        if self.auto_now or self.auto_now_add:
            return self.now()
        return default
    
    def get_python_value(self, value):
        if not value:
            return None
        if isinstance(value, six.string_types):
            try:
                value = datetime.date(*time.strptime(value, '%Y-%m-%d')[:3])
            except ValueError as e:
                raise ValueError('Invalid ISO date %r [%s]' % (value,
                    str(e)))
        return value

    def get_db_value(self, value):
        if value is None:
            return value
        return value.isoformat()


class DateTimeProperty(DateTimeMixin,BaseProperty):

    def __init__(
            self,
            default_value=None,
            required=True,
            validators=[],
            verbose_name=None,
            auto_now=False,
            auto_now_add=False,
            **kwargs):

        super(DateTimeProperty, self).__init__(default_value=default_value,
                                           required=required, validators=validators,
                                           verbose_name=verbose_name, **kwargs)
        self.auto_now = auto_now
        self.auto_now_add = auto_now_add

    def get_default_value(self):
        default = self.default_value
        if self.auto_now or self.auto_now_add:
            return self.now()
        return default

    def get_python_value(self, value):
        if isinstance(value, six.string_types):
            try:
                value = value.split('.', 1)[0] # strip out microseconds
                value = value[0:19] # remove timezone
                value = datetime.datetime.strptime(value, '%Y-%m-%dT%H:%M:%S')
            except ValueError as e:
                raise ValueError('Invalid ISO date/time %r [%s]' %
                        (value, str(e)))
        return value

    def get_db_value(self, value):
        if not value:
            return None
        if self.auto_now:
            value = self.now()

        if value is None:
            return value

        return value.replace(microsecond=0).isoformat() + 'Z'
    
    def now(self):
        return datetime.datetime.utcnow()


import unittest
import datetime
from kev.properties import (BaseProperty, CharProperty, DateProperty,
                            DateTimeProperty, FloatProperty, IntegerProperty,
                            BooleanProperty)
from kev.exceptions import ValidationException


class PropertiesTestCase(unittest.TestCase):

    def test_base_property(self):
        prop = BaseProperty()
        self.assertEqual(None,prop.default_value)
        with self.assertRaises(ValidationException) as vm:
            prop.validate(None,'first_name')
        self.assertEqual(str(vm.exception),
                          'first_name: This value is required')

    def test_base_property_not_required(self):
        prop = BaseProperty(required=False)
        self.assertEqual(prop.required,False)
        prop.validate(None,'first_name')

    def test_base_property_default_value(self):
        prop = BaseProperty(default_value=None)
        self.assertEqual(prop.get_default_value(),None)
        prop = BaseProperty(default_value='Redis')
        self.assertEqual(prop.get_default_value(),'Redis')

    def test_char_property_validators_set(self):
        prop = CharProperty()
        self.assertEqual(2,len(prop.validators))

    def test_char_property_validate(self):
        prop = CharProperty(required=True)
        with self.assertRaises(ValidationException) as vm:
            prop.validate(None, 'first_name')
        self.assertEqual(str(vm.exception),
                          'first_name: This value is required')
        with self.assertRaises(ValidationException) as vm:
            prop.validate(34, 'first_name')
        self.assertEqual(str(vm.exception),
                          'first_name: This value should be a string')

    def test_char_property_validate_default_value(self):
        prop = CharProperty(required=True,default_value='Ben')
        # Make sure that the default value works
        prop.validate(None, 'first_name')
        self.assertEqual(prop.get_default_value(),'Ben')
        prop = CharProperty(required=True,default_value=5)
        with self.assertRaises(ValidationException) as vm:
            prop.validate(None,'first_name')
        self.assertEqual(str(vm.exception),
                          'first_name: This value should be a string')

    def test_integer_property_validators_set(self):
        prop = IntegerProperty()
        self.assertEqual(2, len(prop.validators))

    def test_integer_property_validate(self):
        prop = IntegerProperty(required=True)
        with self.assertRaises(ValidationException) as vm:
            prop.validate(None, 'no_packages')
        self.assertEqual(str(vm.exception),
                          'no_packages: This value is required')
        with self.assertRaises(ValidationException) as vm:
            prop.validate('brains', 'no_packages')
        self.assertEqual(str(vm.exception),
                          'no_packages: This value should be an integer')

    def test_integer_property_validate_default_value(self):
        prop = IntegerProperty(required=True, default_value=5)
        # Make sure that the default value works
        prop.validate(None, 'no_packages')
        self.assertEqual(prop.get_default_value(), 5)
        prop = IntegerProperty(required=True, default_value='Ben')
        with self.assertRaises(ValidationException) as vm:
            prop.validate(None, 'no_packages')
        self.assertEqual(str(vm.exception),
                          'no_packages: This value should be an integer')

    def test_float_property_validators_set(self):
        prop = FloatProperty()
        self.assertEqual(2, len(prop.validators))

    def test_float_property_validate(self):
        prop = FloatProperty(required=True)
        with self.assertRaises(ValidationException) as vm:
            prop.validate(None, 'no_packages')
        self.assertEqual(str(vm.exception),
                          'no_packages: This value is required')
        with self.assertRaises(ValidationException) as vm:
            prop.validate('brains', 'no_packages')
        self.assertEqual(str(vm.exception),
                          'no_packages: This value should be a float.')

    def test_float_property_validate_default_value(self):
        prop = FloatProperty(required=True, default_value=5.5)
        # Make sure that the default value works
        prop.validate(None, 'no_packages')
        self.assertEqual(prop.get_default_value(), 5.5)
        prop = FloatProperty(required=True, default_value=5)
        with self.assertRaises(ValidationException) as vm:
            prop.validate(None, 'no_packages')
        self.assertEqual(str(vm.exception),
                          'no_packages: This value should be a float.')

    def test_date_property_validators_set(self):
        prop = DateProperty()
        self.assertEqual(2, len(prop.validators))

    def test_date_property_validate(self):
        prop = DateProperty(required=True)
        with self.assertRaises(ValidationException) as vm:
            prop.validate(None, 'no_packages')
        self.assertEqual(str(vm.exception),
                          'no_packages: This value is required')
        with self.assertRaises(ValidationException) as vm:
            prop.validate('brains', 'no_packages')
        self.assertEqual(str(vm.exception),
                          'no_packages: This value should be a valid date object.')

    def test_date_property_validate_auto_now(self):
        prop = DateProperty(required=True, auto_now=True)
        # Make sure that the default value works
        prop.validate(None, 'no_packages')
        self.assertEqual(prop.get_default_value(), datetime.date.today())

    def test_date_property_validate_auto_now_add(self):
        prop = DateProperty(required=True, auto_now_add=True)
        # Make sure that the default value works
        prop.validate(None, 'no_packages')
        self.assertEqual(prop.get_default_value(), datetime.date.today())

    def test_date_property_validate_default_value(self):
        prop = DateProperty(required=True, default_value=datetime.date.today())
        # Make sure that the default value works
        prop.validate(None, 'no_packages')
        self.assertEqual(prop.get_default_value(), datetime.date.today())

    def test_datetime_property_validate(self):
        prop = DateTimeProperty(required=True)
        with self.assertRaises(ValidationException) as vm:
            prop.validate(None, 'no_packages')
        self.assertEqual(str(vm.exception),
                          'no_packages: This value is required')
        with self.assertRaises(ValidationException) as vm:
            prop.validate('brains', 'no_packages')
        self.assertEqual(str(vm.exception),
                          'no_packages: This value should be a valid datetime object.')

    def test_datetime_property_validate_default_value(self):
        prop = DateTimeProperty(required=True, default_value=datetime.datetime.now())
        # Make sure that the default value works
        prop.validate(None, 'no_packages')
        self.assertEqual(type(prop.get_default_value()), datetime.datetime)

    def test_datetime_property_validate_auto_now(self):
        prop = DateTimeProperty(requirYYed=True, auto_now=True)
        # Make sure that the default value works
        prop.validate(None, 'no_packages')
        self.assertEqual(type(prop.get_default_value()), datetime.datetime)

    def test_datetime_property_validate_auto_now_add(self):
        prop = DateTimeProperty(required=True, auto_now_add=True)
        # Make sure that the default value works
        prop.validate(None, 'no_packages')
        self.assertEqual(type(prop.get_default_value()), datetime.datetime)

    def test_boolean_property_validate(self):
        prop = BooleanProperty(required=True)
        with self.assertRaises(ValidationException) as vm:
            prop.validate(None, 'is_staff')
        self.assertEqual(str(vm.exception),
                          'is_staff: This value should be True or False.')
        with self.assertRaises(ValidationException) as vm:
            prop.validate('brains', 'is_staff')
        self.assertEqual(str(vm.exception),
                          'is_staff: This value should be True or False.')

    def test_boolean_property_validate_default_value(self):
        prop = BooleanProperty(default_value=True)
        # Make sure that the default value works
        prop.validate(None, 'is_active')
        self.assertEqual(type(prop.get_default_value()), type(True))


if __name__ == '__main__':
    unittest.main()


import unittest

from kev.utils import import_util, import_mod, get_doc_type
from kev.document import Document
from kev.properties import CharProperty

class Frog(Document):
    name = CharProperty()


class Dog(Document):
    name = CharProperty()

    class Meta(object):
        doc_type = 'animal'


class UtilTest(unittest.TestCase):

    def test_import_util(self):
        klass = import_util('kev.backends.s3.db.S3DB')
        self.assertEqual('S3DB',klass.__name__)

    def test_import_mod(self):
        mod = import_mod('kev.backends.s3.db')
        self.assertTrue(hasattr(mod,'S3DB'))

    def test_get_doc_type(self):
        a = get_doc_type(Frog)
        self.assertEqual('Frog',a)
        b = get_doc_type(Dog)
        self.assertEqual('animal', b)

if __name__ == '__main__':
    unittest.main()


import kev.utils


class KevHandler(object):

    def __init__(self,databases):
        #Created for get_db method
        self._databases = dict()
        self._connections = dict()
        self._labels = list()

        for db_label,db_info in databases.items():
            db_klass = kev.utils.import_util(db_info.get('backend'))
            self._databases[db_label] = db_klass(**db_info.get('connection'))

    def get_db(self,db_label):
        return self._databases.get(db_label)



class ValidationException(Exception):
    def __init__(self,msg):
        self.error_msg = msg
        
    def __str__(self):
        return self.error_msg


class DocNotFoundError(Exception):
    pass


class QueryIndexError(Exception):
    pass


class QueryError(Exception):
    pass


class ResourceNotFound(Exception):
    pass


class DocSaveError(Exception):
    pass


class ResourceError(Exception):
    pass



import redis

from kev.backends import DocDB
from kev.exceptions import DocNotFoundError



class RedisDB(DocDB):

    db_class = redis.StrictRedis
    backend_id = 'redis'

    def __init__(self,**kwargs):
        self._db = self._indexer = self.db_class(kwargs['host'],port=kwargs['port'])

    def create_pk(self,doc_obj):
        pk = self._indexer.incr('{0}:{1}:id._pk'.format(self.backend_id,doc_obj.get_class_name()))
        doc_obj.set_pk('{0}:{1}:id:{2}'.format(self.backend_id,doc_obj.get_class_name(), pk))
        return doc_obj

    #CRUD Operations

    def save(self,doc_obj):
        doc_obj, doc = self._save(doc_obj)
        pipe = self._db.pipeline()
        pipe.hmset(doc_obj._id, doc)

        pipe = self.add_to_model_set(doc_obj,pipe)
        pipe = self.add_indexes(doc_obj, doc, pipe)
        pipe = self.remove_indexes(doc_obj, pipe)
        pipe.execute()
        # doc_obj._doc = doc_obj.process_doc_kwargs(doc)
        return doc_obj

    def delete(self,doc_obj):
        pipe = self._db.pipeline()
        pipe.delete(doc_obj._doc['_id'])
        pipe = self.remove_from_model_set(doc_obj,pipe)
        doc_obj._index_change_list = doc_obj.get_indexes()
        pipe = self.remove_indexes(doc_obj, pipe)
        pipe.execute()

    def all(self,cls):

        id_list = self._db.smembers('{0}:all'.format(
            cls.get_class_name()))
        pipe = self._db.pipeline()
        for id in id_list:
            pipe.hgetall(id)

        raw_docs = pipe.execute()
        for doc in raw_docs:
            yield cls(**{k.decode(): v.decode() for k, v in doc.items()})

    def get(self, doc_obj, doc_id):

        doc = self._db.hgetall(doc_obj.get_doc_id(doc_id))
        if len(list(doc.keys())) == 0:
            raise DocNotFoundError

        return doc_obj(**{k.decode(): v.decode() for k, v in doc.items()})

    def flush_db(self):
        self._db.flushdb()

    #Indexing Methods

    def add_to_model_set(self, doc_obj, pipeline):
        pipeline.sadd('{0}:all'.format(doc_obj.__class__.__name__.lower()), doc_obj._id)
        return pipeline

    def remove_from_model_set(self, doc_obj, pipeline):
        pipeline.srem('{0}:all'.format(doc_obj.__class__.__name__.lower()), doc_obj._id)
        return pipeline

    def remove_indexes(self, doc_obj, pipeline):
        for index_v in doc_obj._index_change_list:
            pipeline.srem(index_v, doc_obj._id)
        return pipeline

    def add_indexes(self, doc_obj, doc, pipeline):
        index_list = doc_obj.get_indexed_props()
        for prop in index_list:
            index_value = doc.get(prop)
            if index_value:
                pipeline.sadd(doc_obj.get_index_name(prop, index_value),
                              doc_obj._id
                              )
        return pipeline

    def evaluate(self, filters_list, doc_class):
        if len(filters_list) == 1:
            id_list = self._db.smembers(filters_list[0])
        else:
            id_list = self._db.sinter(*filters_list)
        pipe = self._db.pipeline()
        for id in id_list:
            pipe.hgetall(id)
        raw_docs = pipe.execute()
        for doc in raw_docs:
            yield doc_class(**{k.decode(): v.decode() for k, v in doc.items()})



import six

from kev.exceptions import QueryError, ValidationException
from kev.utils import get_doc_type


class DocDB(object):
    db_class = None
    indexer_class = None
    backend_id = None

    def save(self,doc_obj):
        raise NotImplementedError

    def delete(self, doc_obj):
        raise NotImplementedError

    def get(self, doc_obj, doc_id):
        raise NotImplementedError

    def parse_id(self, doc_id):
        try:
            return doc_id.split(':')[3]
        except TypeError:
            return doc_id.decode().split(':')[3]

    def create_pk(self):
        raise NotImplementedError

    def check_unique(self,doc_obj,key,value):
        obj = doc_obj.objects().filter({key:value})
        if len(obj) == 0:
            return True
        if hasattr(doc_obj,'_id') and len(obj) == 1:
            if doc_obj._id == obj[0]._id:
                return True
        raise ValidationException(
                'There is already a {key} with the value of {value}'\
                    .format(key=key,value=value))

    def _save(self,doc_obj):
        doc = doc_obj._doc.copy()
        for key, prop in list(doc_obj._base_properties.items()):
            prop.validate(doc.get(key), key)
            raw_value = prop.get_python_value(doc.get(key))
            if prop.unique:
                self.check_unique(doc_obj,key,raw_value)
            value = prop.get_db_value(raw_value)
            doc[key] = value


        doc['_doc_type'] = get_doc_type(doc_obj.__class__)

        if '_id' not in doc:
            self.create_pk(doc_obj)
            doc['_id'] = doc_obj._id
        return (doc_obj,doc)
